from typing import Any, Callable, Optional, Type, TypeVar, Union

from graia.broadcast.entities.event import BaseEvent
from graia.broadcast.utilles import run_always_await

from graia.application.group import Group, Member
from graia.application.friend import Friend
from . import Interrupt
from graia.application.event.messages import FriendMessage, GroupMessage, TempMessage
from graia.application.message import BotMessage
from graia.application.message.elements.internal import Quote, Source

class GroupMessageInterrupt(Interrupt):
    direct = GroupMessage
    special_group: Optional[Union[Group, int]] = None
    special_member: Optional[Union[Member, int]] = None
    quote_access: Optional[Union[BotMessage, Source]] = None
    custom_judgement: Optional[Callable[[GroupMessage], bool]] = None

    def __init__(self,
        special_group: Optional[Union[Group, int]] = None,
        special_member: Optional[Union[Member, int]] = None,
        quote_access: Optional[Union[BotMessage, Source]] = None,
        custom_judgement: Optional[Callable[[GroupMessage], bool]] = None
    ) -> None:
        self.special_group = special_group
        self.special_member = special_member
        self.quote_access = quote_access
        self.custom_judgement = custom_judgement

    def set_custom_judgement(self, callable_func):
        self.custom_judgement = callable_func
        return callable_func

    def trigger(self, event: GroupMessage) -> Union[None, GroupMessage]:
        if self.special_group:
            if event.sender.group.id != (self.special_group.id if isinstance(self.special_group, Group) else self.special_group):
                return None
        if self.special_member:
            if event.sender.id != (self.special_member.id if isinstance(self.special_member, Member) else self.special_member):
                return None
        if self.quote_access:
            quotes = event.messageChain.get(Quote)
            if not quotes:
                return
            quote: Quote = quotes[0]
            if isinstance(self.quote_access, BotMessage):
                if quote.id != self.quote_access.messageId:
                    return None
            elif isinstance(self.quote_access, Source):
                if quote.id != self.quote_access.id:
                    return None
        if self.custom_judgement:
            if not self.custom_judgement(event):
                return None
        return event

class FriendMessageInterrupt(Interrupt):
    direct = FriendMessage

    special_friend: Optional[Union[Friend, int]] = None
    quote_access: Optional[Union[BotMessage, Source]] = None
    custom_judgement: Optional[Callable[[FriendMessage], bool]] = None

    def __init__(self,
        special_friend: Optional[Union[Friend, int]] = None,
        quote_access: Optional[Union[BotMessage, Source]] = None,
        custom_judgement: Optional[Callable[[FriendMessage], bool]] = None
    ) -> None:
        self.special_friend = special_friend
        self.quote_access = quote_access
        self.custom_judgement = custom_judgement

    def set_custom_judgement(self, callable_func):
        self.custom_judgement = callable_func
        return callable_func

    def trigger(self, event: FriendMessage) -> Union[None, FriendMessage]:
        if self.special_friend:
            if event.sender.id != (self.special_friend.id if isinstance(self.special_friend, Friend) else self.special_friend):
                return
        if self.quote_access:
            quotes = event.messageChain.get(Quote)
            if not quotes:
                return
            quote: Quote = quotes[0]
            if isinstance(self.quote_access, BotMessage):
                if quote.id != self.quote_access.messageId:
                    return None
            elif isinstance(self.quote_access, Source):
                if quote.id != self.quote_access.id:
                    return None
        if self.custom_judgement:
            if not self.custom_judgement(event):
                return None
        return event

class TempMessageInterrupt(Interrupt):
    direct = TempMessage
    special_group: Optional[Union[Group, int]] = None
    special_member: Optional[Union[Member, int]] = None
    quote_access: Optional[Union[BotMessage, Source]] = None
    custom_judgement: Optional[Callable[[TempMessage], bool]] = None

    def __init__(self,
        special_group: Optional[Union[Group, int]] = None,
        special_member: Optional[Union[Member, int]] = None,
        quote_access: Optional[Union[BotMessage, Source]] = None,
        custom_judgement: Optional[Callable[[TempMessage], bool]] = None
    ) -> None:
        self.special_group = special_group
        self.special_member = special_member
        self.quote_access = quote_access
        self.custom_judgement = custom_judgement

    def set_custom_judgement(self, callable_func):
        self.custom_judgement = callable_func
        return callable_func

    def trigger(self, event: TempMessage) -> Union[None, TempMessage]:
        if self.special_group:
            if event.sender.group.id != (self.special_group.id if isinstance(self.special_group, Group) else self.special_group):
                return None
        if self.special_member:
            if event.sender.id != (self.special_member.id if isinstance(self.special_member, Member) else self.special_member):
                return None
        if self.quote_access:
            quotes = event.messageChain.get(Quote)
            if not quotes:
                return
            quote: Quote = quotes[0]
            if isinstance(self.quote_access, BotMessage):
                if quote.id != self.quote_access.messageId:
                    return None
            elif isinstance(self.quote_access, Source):
                if quote.id != self.quote_access.id:
                    return None
        if self.custom_judgement:
            if not self.custom_judgement(event):
                return None
        return event

T = TypeVar("T", BaseEvent, Any)
A = TypeVar("A")

class CustomEventInterrupt(Interrupt):
    def __init__(self,
        event: Type[T],
        checker: Callable[[T], bool],
        transfer: Callable[[T], A],
    ) -> None:
        self.direct = event
        self.checker = checker
        self.transfer = transfer
    
    async def trigger(self, event: T) -> Union[None, A]:
        if await run_always_await(self.checker(event)):
            return await run_always_await(self.transfer(event))


from typing import Any, Type, Union
from pprint import pprint
from graia.broadcast import Broadcast
from graia.application import GraiaMiraiApplication
from graia.broadcast.entities.event import BaseEvent
from abc import ABCMeta, abstractmethod
from graia.broadcast.utilles import run_always_await
from graia.broadcast.entities.signatures import RemoveMe
import asyncio

class Interrupt(metaclass=ABCMeta):
    """内置中断(Interrupt)机制中, 中断项的抽象实现, 需要重写 `trigger` 方法, 该方法接受一个 event 参数,
    该中断会被立刻回收.
    """
    direct: Type[BaseEvent]
    _return_value: Any = None

    def set_return_value(self, value: Any):
        self._return_value = value

    def get_return_value(self):
        return self._return_value

    @abstractmethod
    def trigger(self, event: BaseEvent) -> Union[None, Any]:
        pass

class InterruptControl:
    broadcast: Broadcast
    app: GraiaMiraiApplication

    def __init__(self, broadcast: Broadcast) -> None:
        self.broadcast = broadcast
    
    async def wait(self, interrupt: Interrupt):
        local_event = asyncio.Event()
        listener_callable = self.leader_listener_generator(local_event, interrupt)
        self.broadcast.receiver(interrupt.direct)(listener_callable)
        try:
            await local_event.wait()
        finally: # 删除 Listener
            if not local_event.is_set():
                self.broadcast.removeListener(self.broadcast.getListener(listener_callable))
        return interrupt.get_return_value()

    def leader_listener_generator(self, event_lock: asyncio.Event, interrupt: Interrupt):
        async def inside_listener(event: interrupt.direct):
            try:
                result = await run_always_await(interrupt.trigger(event))
                if result is not None:
                    event_lock.set()
                    interrupt.set_return_value(result)
                    return RemoveMe()
            except:
                import traceback
                traceback.print_exc()
        return inside_listener